import torch
from torch import nn


def save_model(
    model,
    save_dir,
    epoch=0,
    optimizer=None,
    loss=None,
):
    if isinstance(model, nn.DataParallel):
        state_dict = model.module.state_dict()
    else:
        state_dict = model.state_dict()
    if optimizer is None:
        optim_params = None
    else:
        optim_params = optimizer.state_dict()
    torch.save(
        {
            'init_args': model.init_args,
            'epoch': epoch,
            'model_state_dict': state_dict,
            'optimizer_state_dict': optim_params,
            'loss': loss,
        },
        save_dir
    )


def load_model(
    save_dir,
    model_class=None,
    model=None,
    optimizer=None,
    train=False,
):
    # from .model_dict import MODEL_DICT
    checkpoint = torch.load(save_dir)
    if model is None:
        init_args = checkpoint['init_args']
        assert model_class is not None
        model = model_class(**init_args)
        model.load_state_dict( 
            checkpoint['model_state_dict'], 
        )
    
    elif isinstance(model, nn.DataParallel):
        state_dict = checkpoint['model_state_dict']
        from collections import OrderedDict
        new_state_dict = OrderedDict()

        for k, v in state_dict.items():
            if 'module' not in k:
                k = 'module.' + k
            else:
                k = k.replace('features.module.', 'module.features.')
            new_state_dict[k] = v
        model.load_state_dict(new_state_dict)
    else:
        model.load_state_dict( 
            checkpoint['model_state_dict'], 
        )
    
    if optimizer is not None:
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
    epoch = checkpoint['epoch']
    loss = checkpoint['loss']

    if train:
        model.train()
    else:
        model.eval()

    return model, optimizer, epoch, loss


# shape

import os
import subprocess
from copy import deepcopy

from rdkit import Chem
from rdkit.Chem import AllChem
from pyshapeit import AlignMol
import multiprocess as mp

from rdkit import RDLogger

from jupyfuncs.pbar import tqdm
from jupyfuncs.norm import Normalizer

# from rdkit.Chem.Draw import IPythonConsole
from rdkit.Chem import PyMol
from rdkit.Chem.Subshape import SubshapeBuilder, SubshapeObjects
from PIL import ImageFile
ImageFile.LOAD_TRUNCATED_IMAGES = True

lg = RDLogger.logger()
lg.setLevel(4)

__all__ = [
    'get_mols_from_smi',
    'get_aligned_mol',
    'get_aligned_sdf',
    'show_alignment',
    'pymol_running'
]


def get_mols_from_smi(probe_smifile):
    mols = []
    with open(probe_smifile) as f:
        for line in f.readlines():
            smi = line.strip()
            mol = None
            try:
                mol = Chem.MolFromSmiles(smi)
            except Exception as e:
                print(e)
            if mol:
                mols.append(mol)
    return mols


def get_aligned_mol(
    ref_mol, probe_mol, num_confs, num_cpu
):

    mol1 = ref_mol
    mol1.SetProp('_Name', 'ref')

    AllChem.EmbedMultipleConfs(
        probe_mol,
        numConfs=num_confs,
        numThreads=num_cpu
    )

    score = 0
    # conf_id = -1
    
    aligned_mol = deepcopy(probe_mol)

    for i in range(probe_mol.GetNumConformers()):

        mol2 = Chem.MolFromMolBlock(
            Chem.MolToMolBlock(probe_mol, confId=i)
        )
        mol2.SetProp('_Name', 'probe')

        sim_score = AlignMol(mol1, mol2)
        if sim_score > score:
            score = sim_score
            aligned_mol = deepcopy(mol2)
#     pbar.update(1)

    return aligned_mol


def get_aligned_mol_mp(
    config
):
    return get_aligned_mol(*config)


def gen_configs(ref_mol, mols, num_confs, num_cpu):
    configs = []
    for probe_mol in mols:
        configs.append((ref_mol, probe_mol, num_confs, num_cpu))
    return configs


def get_aligned_sdf(
    ref_sdf: str,
    probe_smifile: str,
    num_confs=150,
    num_cpu=5,
    num_workers=10,
    output_sdf=None,
    print_info=True,
    norm_mol=True
):
    ref_sdf = os.path.abspath(ref_sdf)
    ref_mol = Chem.SDMolSupplier(ref_sdf)[0]
    if not output_sdf:
        output_sdf = os.path.abspath(probe_smifile) + '.sdf'
    else:
        output_sdf = os.path.abspath(output_sdf)

    mols = get_mols_from_smi(probe_smifile)

    configs = gen_configs(
        ref_mol, mols, num_confs=num_confs, num_cpu=num_cpu
    )
    
    pool = mp.Pool(num_workers)
    aligned_mols = list(
        tqdm(
            pool.imap(get_aligned_mol_mp, configs),
            total=len(mols),
            desc='All mols'
        )
    )
    if norm_mol:
        normer = Normalizer()
    sdwriter = Chem.SDWriter(output_sdf) 
    for mol in aligned_mols:
        if norm_mol:
            mol = normer(mol)
        sdwriter.write(mol)
        sdwriter.flush()
    sdwriter.close()
    return output_sdf

    # out_aligned = output_sdf + 'ali.sdf'
    # score_file = output_sdf + 'score.csv'
    
    # command = f'shape-it -r {ref_sdf} -d {output_sdf} -o {out_aligned} -s {score_file}' 
    # out_info = subprocess.getoutput(command)
    # if print_info:
    #     print(out_info)
    
    # if norm_mol:
    #     fix_path = out_aligned + 'fix.sdf'
    #     mols = Chem.SDMolSupplier(out_aligned)
    #     sdwriter = Chem.SDWriter(fix_path)
    #     for mol in mols:
    #         mol = normer(mol)
    #         sdwriter.write(mol)
    #         sdwriter.flush()
    #     sdwriter.close()
    #     return fix_path
    # else:
    #     return out_aligned
    
    # shape-it -r ref_sdf  -d output_sdf -o out_aligned -s score_file


def show_alignment(
    ref_mol,
    probe_mol,
    gen_confs,
    num_confs=200,
    num_cpu=5,
):
    # should install pymol-open-source
    if not pymol_running():
        subprocess.Popen(['pymol', '-cKRQ'])  

    if isinstance(ref_mol, Chem.Mol):
        mol1 = ref_mol
    elif isinstance(ref_mol, str) and ref_mol.endswith('.sdf'):
        mol1 = Chem.SDMolSupplier(ref_mol)[0]
    
    if isinstance(probe_mol, Chem.Mol):
        mol2 = probe_mol
    elif isinstance(probe_mol, str) and probe_mol.endswith('.sdf'):
        mol2 = Chem.SDMolSupplier(probe_mol)[0]
    else:
        mol2 = Chem.MolFromSmiles(probe_mol)
    
    if gen_confs:
        AllChem.EmbedMultipleConfs(
            mol2,
            numConfs=num_confs,
            numThreads=num_cpu
        )

    score = 0
    for i in range(mol2.GetNumConformers()):

        probe_mol = Chem.MolFromMolBlock(
            Chem.MolToMolBlock(mol2, confId=i)
        )

        sim_score = AlignMol(mol1, probe_mol)
        if sim_score > score:
            score = sim_score
            probe = deepcopy(probe_mol)

    mol1.SetProp('_Name', 'ref')
    probe.SetProp('_Name', 'probe')

    AllChem.CanonicalizeConformer(mol1.GetConformer())
    AllChem.CanonicalizeConformer(probe.GetConformer())

    builder = SubshapeBuilder.SubshapeBuilder()
    builder.gridDims = (20., 20., 10)
    builder.gridSpacing = 0.5
    builder.winRad = 4.

    refShape = builder.GenerateSubshapeShape(mol1)
    probeShape = builder.GenerateSubshapeShape(probe)

    v = PyMol.MolViewer()

    score = AlignMol(mol1, probe)

    v.DeleteAll()

    v.ShowMol(mol1, name='ref', showOnly=False)
    SubshapeObjects.DisplaySubshape(v, refShape, 'ref_Shape')
    v.server.do('set transparency=0.5')

    v.ShowMol(probe, name='probe', showOnly=False)
    SubshapeObjects.DisplaySubshape(v, probeShape, 'prob_Shape')
    v.server.do('set transparency=0.5')

    return v.GetPNG()


def pymol_running() -> bool:
    out_info = subprocess.getoutput('ps aux | grep pymol')
    if '-cKRQ' in out_info:
        return True
    else:
        return False


import psycopg
from psycopg import sql


def connect_by_infofile(info_file: str) -> psycopg.Connection:
    """Create a postgres connection

    Args:
        info_file (str): 
            the path of the connection info like
            host=127.0.0.1 dbname=dbname port=5432 user=postgres password=lala

    Returns:
        psycopg.Connection: 
            the connection instance should be closed after committing.
    """
    conn = psycopg.connect(
        open(info_file).readline()
    )
    return conn


def get_item_by_idx(
    idx: int,
    info_file: str,
    by: str = id,
    table: str = 'reaction_id'
):

    conn = connect_by_infofile(
        info_file
    )

    query_name = str(idx)
    query = sql.SQL(
        "select reaction_id from {table} where {by} = %s"
    ).format(
        table=sql.Identifier(table),
        by=sql.Identifier(by)
    )
    cur = conn.execute(query, [query_name]).fetchone()
    return cur[0]


import sys
from os import path as osp
from IPython.core.display import HTML

__all__ = [
    'in_jupyter',
    'tqdm',
    'trange',
    'tnrange',
    'NO_WHITE',
]


NO_WHITE = HTML("""
    <style>
    .jp-OutputArea-prompt:empty {
    padding: 0;
    border: 0;
    }
    </style>
    """)


def in_jupyter():

    which = True if 'ipykernel_launcher.py' in sys.argv[0] else False
    return which

def in_docker():
    return osp.exists('/.dockerenv')


if in_jupyter():
    from tqdm.notebook import tqdm
    from tqdm.notebook import trange
    from tqdm.notebook import tnrange
else:
    from tqdm import tqdm
    from tqdm import trange
    from tqdm import tnrange


"""
Conformer Prediction for classification Task

As we only consider the predicted values as input, 
we do not differentiate between transductive and inductive conformers
"""

import numpy as np
import pandas as pd


class CpClassfier:
    def __init__(self):
        self.cal_data = None
        self.class_num = 0

    def fit_with_data(self, cal_proba, cal_y, class_num=0):
        """
        :parm cal_proba: numpy array of shape [n_samples, n_classes]
                        predicted probability of calibration set
        :parm cal_y: numpy array of shape [n_samples,]
                        true label of calibration set
        """

        if class_num <= 0:
            print("Get class number for input data.")
            self.class_num = cal_proba.shape[1]
        else:
            self.class_num = class_num

        cal_df = pd.DataFrame(cal_proba)
        cal_df.columns = ["class_%d"%i for i in range(self.class_num)]
        cal_df["true_label"] = list(cal_y)
        self.cal_data = cal_df
     
    def fit_with_model(self, func):
        #TODO: besides calibration data, we can also use our model
        pass

    def predict_with_proba(self, X_proba):
        """
        :parm X_proba: numpy array of shape [n_samples, n_classes]
                        predicted probabilities of calibration set
        """
        cp_proba = []
        class_lsts = [sorted(self.cal_data[self.cal_data["true_label"] == i]["class_%d"%i]) \
                        for i in range(self.class_num)]
        proba_lsts = [X_proba[:, i] for i in range(self.class_num)]
        for c_lst, p_lst in zip(class_lsts, proba_lsts):
            c_proba = np.searchsorted(c_lst, p_lst, side='left')/len(c_lst)
            cp_proba.append(c_proba)

        self.cp_P = np.array(cp_proba).T
        return self.cp_P



from os import path as osp
import typing as t
from typing_extensions import Literal

import seaborn as sn
import sklearn
import matplotlib.pyplot as plt
import matplotlib
import numpy as np
import pandas as pd

from .glob import get_num_lines, splitted_strs_from_line 

cm = matplotlib.cm.get_cmap('tab20')
colors = cm.colors
LABEL = Literal[
    'training_size',
    'episode_id',
]


def accuracies_heat(y_true, y_pred, num_tasks):
    assert len(y_true) == len(y_pred)
    cm = sklearn.metrics.confusion_matrix(
        y_true, y_pred, normalize='true'
    )
    df_cm = pd.DataFrame(
        cm, range(num_tasks), range(num_tasks)
    )
    plt.figure(figsize=(10, 10))
    sn.set(font_scale=1.4) 
    sn.heatmap(df_cm, annot=True, annot_kws={"size": 16}, fmt='.2f')


def get_metrics_curves(
    base_dir,
    ckpts,
    num_points,
    title="Metric Curve",
    metric='accuracy',
    log_file='metrics.log',
    label: LABEL = 'training_size',
    save_dir: str = None,
    figsize=(10, 6)
):
    if not save_dir:
        save_dir = osp.join(base_dir, 'metrics_curves.png')
    data_dict = {}
    for ckpt in ckpts:
        log = osp.join(
            base_dir,
            ckpt,
            log_file
        )
        if not osp.exists(log):
            print(f"WARNING: no log file for {ckpt}")
            continue
        data_dict[ckpt] = []
        data_idx = 0
        for line_id in range(get_num_lines(log)):
            line = splitted_strs_from_line(log, line_id)
            if len(line) == 3 and line[1].strip() == metric:
                if label == 'episode_id':
                    x = data_idx
                elif label == 'training_size':
                    x = int(line[0].strip())
                data_dict[ckpt].append(
                    [
                        x,
                        float(line[2].strip())
                    ]
                )
                data_idx += 1
            if line_id >= num_points - 1:
                break
    plt.figure(figsize=figsize, dpi=100)
    # plt.style.use('ggplot')
    plt.title(title)
    for i, (ckpt, points) in enumerate(data_dict.items()):
        points_array = np.array(points).T
        plt.plot(points_array[0], points_array[1], label=ckpt, color=colors[i])
    lg = plt.legend(bbox_to_anchor=(1.2, 1.0), loc='upper right')
    # plt.legend(loc='lower right')
    plt.xlabel(
        label
    )
    plt.ylabel(metric)
    plt.grid(True)
    plt.savefig(
        save_dir,
        format='png', 
        bbox_extra_artists=(lg,), 
        bbox_inches='tight'
    )
    plt.show()


def get_means_vars(
    log_file: str,
    indices: t.List,
    mode: str,
    nears_each: int,
) -> t.List[t.List[int]]:
    
    mean_s, var_s = [], []
    num_points = len(indices)

    nears_lists = []
    if mode == 'id':

        for index in indices:
            nears = []
            nears.extend(list(range(
                index - nears_each, index + 1 + nears_each
            ))) 
            nears_lists.append(nears)
        
        for nears in nears_lists:
            mean_s.append(np.mean([
                float(splitted_strs_from_line(log_file, line_id)[2])
                for line_id in nears
            ]))
            var_s.append((np.std([
                float(splitted_strs_from_line(log_file, line_id)[2])
                for line_id in nears
            ])))

    elif mode == 'value':
        datas = [
            splitted_strs_from_line(log_file, line_id)
            for line_id in range(get_num_lines(log_file))
        ]
        training_sizes = [[int(data[0]) for data in datas]]
        values = np.array([float(data[2]) for data in datas])

        training_sizes = np.repeat(training_sizes, num_points, 0).T
        diffs = training_sizes - indices

        true_indices = np.argmin(np.abs(diffs), 0)

        true_indices_list = [
            list(range(
                index - nears_each, index + 1 + nears_each
            ))
            for index in true_indices
        ]
        mean_s = [
            np.mean(values[indices])
            for indices in true_indices_list
        ]
        var_s = [
            np.std(values[indices])
            for indices in true_indices_list
        ]
        var_s = np.array(var_s) / np.sqrt(num_points)
 
    return mean_s, var_s 


def get_metrics_bars(
    base_dir,
    ckpts,
    title="Metric Bars",
    training_sizes: t.List = [],
    episide_ids: t.List = [],
    nears_each: int = 5,
    pretrained_num: int = 0,
    x_diff: bool = False,
    metric='accuracy',
    log_file='metrics.log',
    label: LABEL = 'training_size',
    save_dir: str = None,
    figsize=(10, 6),
    bar_ratio=0.8,
    minimum=0.0,
    maximum=1.0
):

    if not save_dir:
        save_dir = osp.join(base_dir, 'metrics_bars.png')
    
    x_labels, num_points = [], 0
    if label == 'training_size':
        num_points = len(training_sizes)
        x_labels = training_sizes
        mode = 'value'
    elif label == 'episode_id':
        num_points = len(episide_ids)
        x_labels = episide_ids
        mode = 'id'
    
    x = np.arange(num_points)
    num_strategies = len(ckpts)
    total_width = bar_ratio
    width = total_width / num_strategies  
    x = x - (total_width - width) / 2
        
    if not save_dir:
        save_dir = osp.join(base_dir, 'metrics.png')

    data_dict = {}
    for ckpt in ckpts:

        log = osp.join(
            base_dir,
            ckpt,
            log_file
        )
        if not osp.exists(log):
            print(f"WARNING: no log file for {ckpt}")
            continue
        
        # print(x_labels)
        mean_s, var_s = get_means_vars(
            log_file=log,
            indices=x_labels,
            mode=mode,
            nears_each=nears_each
        )
        data_dict[ckpt] = (mean_s, var_s)
    
    if x_diff:
        x_labels = np.array(x_labels, dtype=np.int) - pretrained_num
 
    plt.figure(figsize=figsize, dpi=100)
    plt.title(title)
    ax = plt.gca()
    ax.set_ylim([minimum, maximum])
    
    for point_idx, (ckpt, datas) in enumerate(data_dict.items()):
        mean_s, var_s = datas
        plt.bar(
            x + width * point_idx,
            mean_s, 
            width=width,
            yerr=var_s,
            tick_label=x_labels,
            label=ckpt,
            color=colors[point_idx]
        )
    
    lg = plt.legend(bbox_to_anchor=(1.2, 1.0), loc='upper right')
    # plt.legend(loc='lower right')
    plt.xlabel(
        label
    )
    plt.ylabel(metric)
    plt.grid(True)
 
    plt.savefig(
        save_dir,
        format='png', 
        bbox_extra_artists=(lg,), 
        bbox_inches='tight'
    )
    plt.show()



# import json
import typing as t
# from tap import Tap


class MolNode(object):
    def __init__(
        self,
        id_type: str = "CAS",
        id_value: str = "111-222-333",
        smiles: str = "",
        routes: str = ""
    ) -> None:
        self.id_type = id_type
        self.id_value = id_value
        self.smiles = smiles
        self.routes = routes
        
        self.parse_dict()

        # self.info_dict = {
        #     "id_type": self.id_type,
        #     "id_value": self.id_value,
        #     "smiles": self.smiles,
        #     "routes": self.routes
        # }
    
    def parse_dict(self):
        routes = []
        for route in self.routes:
            parse_dict = getattr(route, 'parse_dict', None)
            if parse_dict is not None:
                route.parse_dict()
            # route_dict = route.get_dict()
            if isinstance(route, t.Dict):
                routes.append(route)
            else:
                routes.append(route.get_dict())
                
        self.routes = routes

    def get_dict(self):
        self.parse_dict()
        self.info_dict = {
            "id_type": self.id_type,
            "id_value": self.id_value,
            "smiles": self.smiles,
            "routes": self.routes
        }
        return self.info_dict


class RouteNode(object):
    def __init__(
        self,
        rxn_template: str = "",
        rxn_type: str = "",
        yld: float = 0.0,
        children: t.List[MolNode] = []
    ):
        self.rxn_template = rxn_template
        self.rxn_type = rxn_type
        self.yld = yld
        self.children = children

        self.parse_dict()
    
        # self.info_dict = {
        #     "rxn_template": self.rxn_template,
        #     "rxn_type": rxn_type,
        #     "yield": self.yld,
        #     "children": self.children
        # }
    
    def parse_dict(self):
        children = []
        for mol_node in self.children:
            parse_dict = getattr(mol_node, 'parse_dict', None)
            if parse_dict is not None:
                mol_node.parse_dict()
            if isinstance(mol_node, t.Dict):
                children.append(mol_node)
            else:
                children.append(mol_node.get_dict())
        self.children = children

    def get_dict(self):
        self.parse_dict()
        self.info_dict = {
            "rxn_template": self.rxn_template,
            "rxn_type": self.rxn_type,
            "yield": self.yld,
            "children": self.children
        }
        return self.info_dict



"""UNCERTAINTY SAMPLING

Uncertainty Sampling examples for Active Learning in PyTorch

It contains four Active Learning strategies:
1. Least Confidence Sampling
2. Margin of Confidence Sampling
3. Ratio of Confidence Sampling
4. Entropy-based Sampling

"""
from copy import deepcopy

import numpy as np

__all__ = [
    "get_prob_unc",
]


def least_conf_unc(prob_array: np.ndarray) -> np.ndarray:
    """Least confidence uncertainty

    .. math::
        \phi_{L C}(x)=\left(1-P_{\theta}\left(y^{*} \mid x\right)\right) \times \frac{n}{n-1}

    Args:
        prob_array (np.array): a 1D or 2D array of probabilities
 
    Returns:
        np.ndarray: the uncertainty value(s)
    """
    if prob_array.ndim == 1:
        indices = prob_array.argmax()
    else:
        indices = (
            np.arange(prob_array.shape[0]),
            prob_array.argmax(-1)
        )
    num_labels = prob_array.shape[-1]
    uncs = (1 - prob_array[indices]) * (num_labels / (num_labels - 1))
    return uncs


def margin_conf_unc(prob_array: np.ndarray) -> np.ndarray:
    """The margin confidence uncertainty

    .. math:: 
        \phi_{M C}(x)=1-\left(P_{\theta}\left(y_{1}^{*} \mid x\right)-P_{\theta}\left(y_{2}^{*} \mid x\right)\right)

    Args:
        prob_array (np.array): a 1D or 2D probability array from an NN.  

    Returns:
        np.array: the uncertainty value(s)
    """
    probs = deepcopy(prob_array)
    probs.sort(-1)
    diffs = probs[..., -1] - probs[..., -2]
    return 1 - diffs


def ratio_conf_unc(prob_array: np.ndarray) -> np.ndarray:
    """Ratio based uncertainties

    .. math::
            \phi_{R C}(x)=P_{\theta}\left(y_{2}^{*} \mid x\right) / P_{\theta}\left(y_{1}^{*} \mid x\right)

    Args:
        prob_array (np.array): a 1D or 2D probability array

    Returns:
        np.array: the uncertainty value(s)
    """
    probs = deepcopy(prob_array)
    probs.sort(-1)
    ratio = probs[..., -1] / probs[..., -2]
    return ratio


def entropy_unc(prob_array: np.ndarray) -> np.ndarray:
    """Entropy based uncertainty

    .. math::
        \phi_{E N T}(x)=\frac{-\Sigma_{y} P_{\theta}(y \mid x) \log _{2} P_{\theta}(y \mid x)}{\log _{2}(n)}

    Args:
        prob_array (np.array): a 1D or 2D probability array

    Returns:
        np.array: the uncertainty value(s)
    """
    num_labels = prob_array.shape[-1]
    log_probs = prob_array * np.log2(prob_array)
    
    raw_entropy = 0 - np.sum(log_probs, -1)

    normalized_entropy = raw_entropy / np.log2(num_labels)

    return normalized_entropy


unc_dict = {
    'least': least_conf_unc,
    'margin': margin_conf_unc,
    'ratio': ratio_conf_unc,
    'entropy': entropy_unc,
}


def get_prob_unc(prob_array: np.ndarray, unc: str) -> np.ndarray:
    return unc_dict[unc](prob_array)



# Extract ligand and pdb 

import sys
from prody import *
from rdkit import Chem
from rdkit.Chem import AllChem
from io import StringIO
import pypdb


__all__ = [
    'get_pdb_components',
    'process_ligand',
    'write_pdb',
    'write_sdf'
]

def get_pdb_components(pdb_id):
    """
    Split a protein-ligand pdb into protein and ligand components
    :param pdb_id:
    :return:
    """
    pdb = parsePDB(pdb_id)
    protein = pdb.select('protein')
    ligand = pdb.select('not protein and not water')
    return protein, ligand


def process_ligand(ligand, res_name):
    """
    Add bond orders to a pdb ligand
    1. Select the ligand component with name "res_name"
    2. Get the corresponding SMILES from pypdb
    3. Create a template molecule from the SMILES in step 2
    4. Write the PDB file to a stream
    5. Read the stream into an RDKit molecule
    6. Assign the bond orders from the template from step 3
    :param ligand: ligand as generated by prody
    :param res_name: residue name of ligand to extract
    :return: molecule with bond orders assigned
    """
    output = StringIO()
    sub_mol = ligand.select(f"resname {res_name}")
    chem_desc = pypdb.describe_chemical(f"{res_name}")
    sub_smiles = chem_desc["describeHet"]["ligandInfo"]["ligand"]["smiles"]
    template = AllChem.MolFromSmiles(sub_smiles)
    writePDBStream(output, sub_mol)
    pdb_string = output.getvalue()
    rd_mol = AllChem.MolFromPDBBlock(pdb_string)
    new_mol = AllChem.AssignBondOrdersFromTemplate(template, rd_mol)
    return new_mol


def write_pdb(protein, pdb_name):
    """
    Write a prody protein to a pdb file
    :param protein: protein object from prody
    :param pdb_name: base name for the pdb file
    :return: None
    """
    output_pdb_name = f"{pdb_name}_protein.pdb"
    writePDB(f"{output_pdb_name}", protein)
    print(f"wrote {output_pdb_name}")


def write_sdf(new_mol, pdb_name, res_name):
    """
    Write an RDKit molecule to an SD file
    :param new_mol:
    :param pdb_name:
    :param res_name:
    :return:
    """
    outfile_name = f"{pdb_name}_{res_name}_ligand.sdf"
    writer = Chem.SDWriter(f"{outfile_name}")
    writer.write(new_mol)
    print(f"wrote {outfile_name}")


def main(pdb_name):
    """
    Read Ligand Expo data, split pdb into protein and ligands,
    write protein pdb, write ligand sdf files
    :param pdb_name: id from the pdb, doesn't need to have an extension
    :return:
    """
    protein, ligand = get_pdb_components(pdb_name)
    write_pdb(protein, pdb_name)

    res_name_list = list(set(ligand.getResnames()))
    for res in res_name_list:
        new_mol = process_ligand(ligand, res)
        write_sdf(new_mol, pdb_name, res)


import os
import typing as t
import inspect
import fnmatch
import linecache
import time
import gc
import psutil
from os import path as osp
import pathlib
import sys
import importlib
import subprocess

import multiprocess as mp

__all__ = [
    'recursive_glob',
    'makedirs',
    'get_current_dir',
]


def recursive_glob(treeroot, pattern):
    results = []
    for base, dirs, files in os.walk(treeroot):
        goodfiles = fnmatch.filter(files, pattern)
        results.extend(os.path.join(base, f) for f in goodfiles)
    return results


def makedirs(path: str, isfile: bool = False) -> None:
    """Creates a directory given a path to either a directory or file.
    If a directory is provided, creates that directory. If a file is provided (i.e. :code:`isfile == True`),
    creates the parent directory for that file.


    Args:
        path (str): Path to a directory or file. 
        isfile (bool, optional): Whether the provided path is a directory or file.Defaults to False.
    """    
    if isfile:
        path = os.path.dirname(path)
    if path != '':
        os.makedirs(path, exist_ok=True)


def get_current_dir():
    return os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))


def get_num_lines(file):
    num_lines = subprocess.check_output(
        ['wc', '-l', file]
    ).split()[0]
    return int(num_lines)


def str_from_line(file, line, split=False):
    smi = subprocess.check_output(
        # ['sed','-n', f'{str(i+1)}p', file]
        ["sed", f"{str(line + 1)}q;d", file]
    )
    if isinstance(smi, bytes):
        smi = smi.decode().strip()
    if split:
        if ' ' or '\t' in smi:
            smi = smi.split()[0]
    return smi


def splitted_strs_from_line(
    file: str,
    idx: int
) -> t.List:
    return str_from_line(file, idx).split()


def chunkify_file(
    fname,
    size=1024 * 1024 * 1000,
    skiplines=-1
):
    """
    function to divide a large text file into chunks each having size ~= size so that the chunks are line aligned

    Params : 
        fname : path to the file to be chunked
        size : size of each chink is ~> this
        skiplines : number of lines in the begining to skip, -1 means don't skip any lines
    Returns : 
        start and end position of chunks in Bytes
    """
    chunks = []
    fileEnd = os.path.getsize(fname)
    with open(fname, "rb") as f:
        if(skiplines > 0):
            for i in range(skiplines):
                f.readline()

        chunkEnd = f.tell()
        count = 0
        while True:
            chunkStart = chunkEnd
            f.seek(f.tell() + size, os.SEEK_SET)
            f.readline()  # make this chunk line aligned
            chunkEnd = f.tell()
            chunks.append((chunkStart, chunkEnd - chunkStart, fname))
            count += 1

            if chunkEnd > fileEnd:
                break
    return chunks


def parallel_apply_line_by_line_chunk(chunk_data):
    """
    function to apply a function to each line in a chunk

    Params :
        chunk_data : the data for this chunk 
    Returns :
        list of the non-None results for this chunk
    """
    chunk_start, chunk_size, file_path, func_apply = chunk_data[:4]
    func_args = chunk_data[4:]

    # t1 = time.time()
    chunk_res = []
    with open(file_path, "rb") as f:
        f.seek(chunk_start)
        cont = f.read(chunk_size).decode(encoding='utf-8')
        lines = cont.splitlines()

        for _, line in enumerate(lines):
            ret = func_apply(line, *func_args)
            if(ret != None):
                chunk_res.append(ret)
    return chunk_res


def parallel_apply_line_by_line(
    input_file_path,
    chunk_size_factor,
    num_procs,
    skiplines,
    func_apply,
    func_args,
    fout=None
):
    """
    function to apply a supplied function line by line in parallel

    Params :
        input_file_path : path to input file
        chunk_size_factor : size of 1 chunk in MB
        num_procs : number of parallel processes to spawn, max used is num of available cores - 1
        skiplines : number of top lines to skip while processing
        func_apply : a function which expects a line and outputs None for lines we don't want processed
        func_args : arguments to function func_apply
        fout : do we want to output the processed lines to a file
    Returns :
        list of the non-None results obtained be processing each line
    """
    num_parallel = min(num_procs, psutil.cpu_count()) - 1

    jobs = chunkify_file(input_file_path, 1024 * 1024 * chunk_size_factor, skiplines)

    jobs = [list(x) + [func_apply] + func_args for x in jobs]

    print("Starting the parallel pool for {} jobs ".format(len(jobs)))

    lines_counter = 0

    pool = mp.Pool(num_parallel, maxtasksperchild=1000)  # maxtaskperchild - if not supplied some weird happend and memory blows as the processes keep on lingering

    outputs = []
    for i in range(0, len(jobs), num_parallel):
        print("Chunk start = ", i)
        t1 = time.time()
        chunk_outputs = pool.map(
            parallel_apply_line_by_line_chunk,
            jobs[i: i + num_parallel]
        )

        for i, subl in enumerate(chunk_outputs):
            for x in subl:
                if(fout != None):
                    print(x, file=fout)
                else:
                    outputs.append(x)
                lines_counter += 1
        del(chunk_outputs)
        gc.collect()
        print("All Done in time ", time.time() - t1)

    print("Total lines we have = {}".format(lines_counter))

    pool.close()
    pool.terminate()
    return outputs


def get_func_from_dir(score_dir: str) -> t.Tuple[t.Callable, str]:
    if score_dir.endswith('.py'):
        func_dir = pathlib.Path(score_dir).parent.resolve()
        file_name = pathlib.Path(score_dir).stem
    else:
        func_dir = osp.abspath(score_dir)
        file_name = "main"

    sys.path.append(func_dir)
    module = importlib.import_module(file_name) 
    try:
        mode = module.MODE
    except Exception as _:
        mode = 'batch'
    return module.main, mode 


import typing as t
from collections import defaultdict

import torch
import numpy as np
import pandas as pd
from scipy.spatial.distance import cdist

def spmmsp(
    sp1: torch.sparse.Tensor,
    sp2: torch.sparse.Tensor
) -> torch.sparse.Tensor:
    from torch_sparse import spspmm
    assert sp1.size(-1) == sp2.size(0) and sp1.is_sparse and sp2.is_sparse
    m = sp1.size(0)
    k = sp2.size(0)
    n = sp2.size(-1)
    indices, values = spspmm(
        sp1.indices(), sp1.values(),
        sp2.indices(), sp2.values(),
        m, k, n
    )
    return torch.sparse_coo_tensor(
        indices,
        values,
        torch.Size([m, n])
    )


def label_to_onehot(ls, class_num, missing_label=-1):
    """
    example:
    >>>label_to_onehot([2,3,-1],6,-1)
    array([[ 0.,  0.,  1.,  0.],
       [ 0.,  0.,  0.,  1.],
       [nan, nan, nan, nan]])
    :param ls:
    :param class_num:
    :param missing_label:
    :return:
    """
    if isinstance(ls, torch.Tensor):
        bool_t = ls == missing_label
        clamp_t = torch.clamp(ls, min=0)
        full_tensor = torch.zeros(ls.numel(), class_num)
        full_tensor = full_tensor.scatter_(1, clamp_t.reshape(-1, 1), 1)
        full_tensor[bool_t] = 0
        return full_tensor
    elif isinstance(ls, t.List):
        ls = np.array(ls, dtype=np.int32)
        bool_array = ls == missing_label
        arr = np.zeros((ls.size, ls.max() + 1))
        arr[np.arange(ls.size), ls] = 1
        arr[bool_array] = 0
        return arr
    elif not isinstance(ls, t.Iterable):
        arr = torch.zeros(class_num)
        if ls != missing_label and not np.isnan(ls) and ls is not None:
            arr[int(ls)] = 1
        return arr


def onehot_to_label(tensor):
    if isinstance(tensor, torch.Tensor):
        return torch.argmax(tensor, dim=-1)
    elif isinstance(tensor, np.ndarray):
        return np.argmax(tensor, axis=-1)


def label_to_tensor(
    label,
    num_classes,
    missing_label=-1,
    device=torch.device('cpu')
):
    if isinstance(label, t.List) and not any(label):
        return torch.zeros(num_classes).to(device)
    elif isinstance(label, t.List) and isinstance(label[0], t.Iterable):
        max_length = max([len(_l) for _l in label])
        index = [_l + _l[-1:] * (max_length - len(_l)) for _l in label]
        tensor_list = []
#         tensor = torch.zeros(len(label), num_classes, device=device)
        for _idx in index:
            _tensor = torch.zeros(num_classes).to(device)
            _idx = torch.LongTensor(_idx)
            _tensor = _tensor.scatter(0, _idx, 1)
            tensor_list.append(_tensor)
        
        return torch.vstack(tensor_list).to(device)
    else:
        if label == missing_label or np.isnan(label):
            return torch.zeros(num_classes)
        tensor = torch.zeros(num_classes).to(device)
        tensor[int(label)] = 1
        return tensor


def tensor_to_label(tensor, threshold=0.5):
    label_list, label_dict = [], defaultdict(list)
    labels = (tensor > threshold).nonzero(as_tuple=False)
    for label in labels:
        label_dict[label[0].item()].append(label[1].item())
    for _, label_value in label_dict.items():
        label_list.append(label_value)
    return label_list 


def get_dist_matrix(
    a: np.ndarray, b: np.ndarray
):
    return cdist(a, b)
    # aSumSquare = np.sum(np.square(a), axis=1)
    # bSumSquare = np.sum(np.square(b), axis=1)
    # mul = np.dot(a, b.T)
    # dists = np.sqrt(aSumSquare[:, np.newaxis] + bSumSquare - 2 * mul)
    # return dists


def get_valid_indices(labels):
    if isinstance(labels, torch.Tensor):
        nan_indices = torch.isnan(labels)
        valid_indices = (
            nan_indices == False
        ).nonzero(as_tuple=True)[0]
    else:
        target_pd = pd.array(labels)
        nan_indices = pd.isna(target_pd)
        valid_indices = torch.LongTensor(
            np.where(nan_indices == False)[0]
        )
    return valid_indices


def smooth_max(
    tensor: torch.Tensor,
    inf_k: int = None,
    **kwargs
):
    if inf_k is None:
        inf_k = 10
    max_value = torch.log(
        torch.sum(
            torch.exp(tensor * inf_k),
            **kwargs
        )
    ) / inf_k
    return max_value


def list_df(listA, listB):
    retB = list(set(listA).intersection(set(listB)))

    retC = list(set(listA).union(set(listB)))

    retD = list(set(listB).difference(set(listA)))

    retE = list(set(listA).difference(set(listB)))
    return retB, retC, retD, retE




# Jupyter funcs
import os
import re
import itertools
from copy import deepcopy
# from collections import defaultdict

from rdkit import Chem
from rdkit.Chem.Draw import IPythonConsole
from rdkit.Chem.Draw.IPythonConsole import addMolToView
# from rdkit.Chem import rdDepictor
from rdkit.Chem.Draw import rdMolDraw2D
from IPython.display import SVG
from rdkit.Chem import AllChem
from ipywidgets import (
    interact,
    # interactive,
    fixed,
)
from rdkit.Chem.rdRGroupDecomposition import (
    RGroupDecomposition,
    # RGroupDecompositionParameters,
    # RGroupMatching,
    # RGroupScore,
    # RGroupLabels,
    # RGroupCoreAlignment,
    RGroupLabelling
)
import pandas as pd
from rdkit.Chem import PandasTools
from rdkit.Chem import Draw
from IPython.display import HTML
# from rdkit import rdBase
from IPython.display import display

from rdkit import Chem
from rdkit.Chem import rdmolops
from rdkit.Chem import Draw
from rdkit.Chem.Draw import IPythonConsole
from rdkit.Chem import rdRGroupDecomposition
from rdkit.Chem import rdqueries
from rdkit.Chem import rdDepictor
from rdkit.Chem.Draw import rdMolDraw2D
from rdkit.Chem.MolStandardize import rdMolStandardize
from rdkit import Geometry
rdDepictor.SetPreferCoordGen(True)
import pandas as pd
from PIL import Image as pilImage
from io import BytesIO
from IPython.display import SVG, Image
from ipywidgets import interact
import molvs as mv


IPythonConsole.ipython_useSVG = True
IPythonConsole.molSize = (450, 350)
params = Chem.SubstructMatchParameters()
params.aromaticMatchesConjugated = True 

__all__ = [
    'draw_mol',
    'draw_confs',
    'show_decomp',
    'get_ids_folds',
    'show_pharmacophore',
    'mol_without_indices',
    'norm_colors',
    'drawmol_with_hi',
    'draw_mols_surfs',
]


COLORS = {
    # "Tol" colormap from https://davidmathlogic.com/colorblind
    'tol': [
        (51, 34, 136),
        (17, 119, 51),
        (68, 170, 153),
        (136, 204, 238),
        (221, 204, 119),
        (204, 102, 119),
        (170, 68, 153),
        (136, 34, 85)
    ],
    # "IBM" colormap from https://davidmathlogic.com/colorblind
    'ibm': [
        (100, 143, 255),
        (120, 94, 240),
        (220, 38, 127),
        (254, 97, 0),
        (255, 176, 0)
    ],
    # Okabe_Ito colormap from https://jfly.uni-koeln.de/color/
    'okabe': [
        (230, 159, 0),
        (86, 180, 233),
        (0, 158, 115),
        (240, 228, 66),
        (0, 114, 178),
        (213, 94, 0),
        (204, 121, 167)
    ]
}


# def get_his_for_onemol(mol_sm, pat_sm):
#     atom_ids = []
#     bond_ids = []
#     m = Chem.MolFromSmiles(mol_sm)
#     pt = Chem.MolFromSmiles(pat_sm)
#     hi_id = m.GetSubstructMatches(pt, params=params)
#     if len(m.GetSubstructMatches(pt, params=params)) == 0:
#         Chem.Kekulize(m)       
#         hi_id = m.GetSubstructMatches(pt, params=params) 
#     if len(hi_id) == 0:
#         return
#     atom_ids.append(itertools.chain.from_iterable(hi_id))


# def get_match_his(mol_sms, pat_sms):
#     highlightatoms = defaultdict(list)
#     highlightbonds = defaultdict(list)
#     for i in range(len(df)):
#         try:
#             mm = df.iloc[i, 0][2:-2]
#             pm = df.iloc[i, 2]
#             m = Chem.MolFromSmiles(mm) 
#             pt = Chem.MolFromSmiles(pm)
#             hi_id = m.GetSubstructMatches(pt, params=params)
#             if len(m.GetSubstructMatches(pt, params=params)) == 0:
#                 Chem.Kekulize(m)       
#                 hi_id = m.GetSubstructMatches(pt, params=params)
#             mols.append(m)
#             hi_ids.append(hi_id)
#         except:
#             pass
#         pass


def norm_colors(colors=COLORS):
    colors = deepcopy(COLORS)
    for k, v in colors.items():
        for i, color in enumerate(v):
            colors[k][i] = tuple(y / 255 for y in color)
    return colors


def drawmol_with_hi(
    mol,
    legend,
    atom_hi_dict,
    bond_hi_dict,
    atomrads_dict,
    widthmults_dict,
    width=350,
    height=200,
):
    d2d = rdMolDraw2D.MolDraw2DCairo(width, height)
    d2d.ClearDrawing()
    d2d.DrawMoleculeWithHighlights(
        mol, legend, 
        atom_hi_dict,
        bond_hi_dict, 
        atomrads_dict,
        widthmults_dict
    )
    d2d.FinishDrawing()
    png = d2d.GetDrawingText()
    return png


def show_atom_number(mol, label='atomNote'):
    new_mol = deepcopy(mol)
    for atom in new_mol.GetAtoms():
        atom.SetProp(label, str(atom.GetIdx()))
    return new_mol


def moltosvg(mol, molSize=(500, 500), kekulize=True):
    mc = mol
    drawer = rdMolDraw2D.MolDraw2DSVG(molSize[0], molSize[1])
    drawer.DrawMolecule(mc)
    drawer.FinishDrawing()
    svg = drawer.GetDrawingText()
    return svg.replace('svg:', '')


def draw_mol(mol):
    return SVG(moltosvg(show_atom_number(mol)))


def drawit(m, p, confId=-1):
    mb = Chem.MolToMolBlock(m, confId=confId)
    p.removeAllModels()
    p.addModel(mb, 'sdf')
    p.setStyle({'stick': {}})
    p.setBackgroundColor('0xeeeeee')
    p.zoomTo()
    return p.show()


def draw_confs(m):
    import py3Dmol
    p = py3Dmol.view(width=500, height=500)
    return interact(drawit,
                    m=fixed(m),
                    p=fixed(p),
                    confId=(0, m.GetNumConformers() - 1))


def do_decomp(mols, cores, options):
    options.rgroupLabelling = RGroupLabelling.AtomMap
    decomp = RGroupDecomposition(cores, options)
    for mol in mols:
        decomp.Add(mol)
    decomp.Process()
    return decomp


def show_decomp(mols, cores, options, item=False):
    decomp = do_decomp(mols, cores, options)
    if item:
        rows = decomp.GetRGroupsAsRows()
        items = [
            '{}:{}'.format(
                group, Chem.MolToSmiles(row[group])
            )
            for row in rows for group in row
        ]
        return ' '.join(items)
    else:
        cols = decomp.GetRGroupsAsColumns()
        cols['mol'] = mols
        cols['input core'] = cores[0]
        df = pd.DataFrame(cols)
        PandasTools.ChangeMoleculeRendering(df)
        return HTML(df.to_html())


def get_ids_folds(id_list, num_folds, need_shuffle=False):
    if need_shuffle:
        from random import shuffle
        shuffle(id_list)
    num_ids = len(id_list)
    assert num_ids >= num_folds
    
    num_each_fold = int(num_ids / num_folds)
    
    blocks = []
    
    for i in range(num_folds):
        start = num_each_fold * i
        end = start + num_each_fold
        if end > num_ids - 1:
            end = num_ids - 1
        
        blocks.append(id_list[start: end])
    
    id_blocks = []
    for i in range(num_folds):
        id_blocks.append(
            (list(itertools.chain.from_iterable([blocks[j] for j in range(num_folds) if j != i])),
             blocks[i])
        )
        
    return id_blocks


keep = ["Donor", "Acceptor", "Aromatic", "Hydrophobe", "LumpedHydrophobe"]


def show_pharmacophore(
    sdf_path,
    keep=keep,
    fdf_dir=os.path.join(
        os.path.dirname(__file__),
        'defined_BaseFeatures.fdef'
    )
):
    template_mol = [m for m in Chem.SDMolSupplier(sdf_path)][0]
    fdef = AllChem.BuildFeatureFactory(
        fdf_dir
    )
    prob_feats = fdef.GetFeaturesForMol(template_mol)
    prob_feats = [f for f in prob_feats if f.GetFamily() in keep]
    # prob_points = [list(x.GetPos()) for x in prob_feats]

    for i, feat in enumerate(prob_feats):
        atomids = feat.GetAtomIds()
        print(
            "pharamcophore index:{0}; feature:{1}; type:{2}; atom id:{3}".format(
                i, 
                feat.GetFamily(), 
                feat.GetType(),
                atomids
            )
        )
        display(
            Draw.MolToImage(
                template_mol,
                highlightAtoms=list(atomids),
                highlightColor=[0, 1, 0],
                useSVG=True
            )
        )


def mol_without_indices( 
    mol_input: Chem.Mol, 
    remove_indices=[], 
    keep_properties=[] 
): 
     
    atom_list, bond_list, idx_map = [], [], {}  # idx_map: {old: new} 
    for atom in mol_input.GetAtoms(): 
         
        props = {} 
        for property_name in keep_properties: 
            if property_name in atom.GetPropsAsDict(): 
                props[property_name] = atom.GetPropsAsDict()[property_name] 
        symbol = atom.GetSymbol() 
         
        if symbol.startswith('*'): 
            atom_symbol = '*' 
            props['molAtomMapNumber'] = atom.GetAtomMapNum() 
        elif symbol.startswith('R'): 
            atom_symbol = '*' 
            if len(symbol) > 1: 
                atom_map_num = int(symbol[1:]) 
            else: 
                atom_map_num = atom.GetAtomMapNum() 
            props['dummyLabel'] = 'R' + str(atom_map_num) 
            props['_MolFileRLabel'] = str(atom_map_num) 
            props['molAtomMapNumber'] = atom_map_num 
             
        else: 
            atom_symbol = symbol 
        atom_list.append( 
            ( 
                atom_symbol, 
                atom.GetFormalCharge(), 
                atom.GetNumExplicitHs(), 
                props 
            ) 
        ) 
    for bond in mol_input.GetBonds(): 
        bond_list.append( 
            ( 
                bond.GetBeginAtomIdx(), 
                bond.GetEndAtomIdx(), 
                bond.GetBondType() 
            ) 
        ) 
    mol = Chem.RWMol(Chem.Mol()) 
     
    new_idx = 0 
    for atom_index, atom_info in enumerate(atom_list): 
        if atom_index not in remove_indices: 
            atom = Chem.Atom(atom_info[0]) 
            atom.SetFormalCharge(atom_info[1]) 
            atom.SetNumExplicitHs(atom_info[2]) 
             
            for property_name in atom_info[3]: 
                if isinstance(atom_info[3][property_name], str): 
                    atom.SetProp(property_name, atom_info[3][property_name]) 
                elif isinstance(atom_info[3][property_name], int): 
                    atom.SetIntProp(property_name, atom_info[3][property_name]) 
            mol.AddAtom(atom) 
            idx_map[atom_index] = new_idx 
            new_idx += 1 
    for bond_info in bond_list: 
        if ( 
            bond_info[0] not in remove_indices 
            and bond_info[1] not in remove_indices 
        ): 
            mol.AddBond( 
                idx_map[bond_info[0]], 
                idx_map[bond_info[1]], 
                bond_info[2] 
            ) 
        else: 
            one_in = False 
            if ( 
                (bond_info[0] in remove_indices) 
                and (bond_info[1] not in remove_indices) 
            ): 
                keep_index = bond_info[1] 
                # remove_index = bond_info[0] 
                one_in = True 
            elif ( 
                (bond_info[1] in remove_indices) 
                and (bond_info[0] not in remove_indices) 
            ): 
                keep_index = bond_info[0] 
                # remove_index = bond_info[1] 
                one_in = True 
            if one_in:  
                if atom_list[keep_index][0] == 'N': 
                    old_num_explicit_Hs = mol.GetAtomWithIdx( 
                        idx_map[keep_index] 
                    ).GetNumExplicitHs() 

                    mol.GetAtomWithIdx(idx_map[keep_index]).SetNumExplicitHs( 
                        old_num_explicit_Hs + 1 
                    ) 
    mol = Chem.Mol(mol) 
    return mol


def draw_mols_surfs(
    mols,
    width=400,
    height=400,
    surface=True,
    surface_opacity=0.5
):
    import py3Dmol

    view = py3Dmol.view(width=width, height=height)
    view.setBackgroundColor('0xeeeeee')
    view.removeAllModels()
    for mol in mols:
        addMolToView(mol, view)
    if surface:
        view.addSurface(
            py3Dmol.SAS,
            {'opacity': surface_opacity}
        )
    view.zoomTo()
    return view.show()


def draw_rxn(
    rxn_smiles,
    use_smiles: bool = True,
):
    rxn = AllChem.ReactionFromSmarts(rxn_smiles, useSmiles=use_smiles)
    d2d = Draw.MolDraw2DCairo(2000, 500)
    d2d.DrawReaction(rxn, highlightByReactant=True)
    png = d2d.GetDrawingText()
    display(Image(png))


def react(rxn_smarts, reagents):
    try:
        rxn = AllChem.ReactionFromSmarts(rxn_smarts)
        # n_reactants = rxn.GetNumReactantTemplates()
        products = rxn.RunReactants([
            Chem.MolFromSmiles(smi) for smi in reagents
        ])
        return products
    except Exception as e:
        print(e)
        return []


def match_pattern(mol, patt):
    if mol:
        return mol.HasSubstructMatch(patt)
    else:
        return False


def split_rxn_smiles(smi):
    try:
        reagents1, reagents2, products = smi.split('>')
        if len(reagents2) > 0:
            reagents = '.'.join([reagents1, reagents2])
        else:
            reagents = reagents1
        return reagents, products
    except Exception as e:
        print(e)
        return '', ''


def find_mprod(rxn_smi):
    # ref: https://github.com/LiamWilbraham/uspto-analysis/blob/master/reaction-stats-uspto.ipynb
    rxn_smarts = '[C:1](=[O:2])-[OD1].[N!H0:3]>>[C:1](=[O:2])[N:3]'
    patt_acid = Chem.MolFromSmarts('[CX3](=O)[OX2H1]')
    patt_amine = Chem.MolFromSmarts('[N;H3,H2,H1]')  # ammonia or primary/secondary amine
    
    products = split_rxn_smiles(rxn_smi)[1].split('.')
        
    reactants = [r for r in split_rxn_smiles(rxn_smi)[0].split('.')]
    cooh = [
        r
        for r in reactants
        if match_pattern(Chem.MolFromSmiles(r), patt_acid)
    ]
    
    cooh = [re.sub('@', '', i) for i in cooh]
    
    amine = [
        r for r in reactants
        if match_pattern(Chem.MolFromSmiles(r), patt_amine)
    ]
    amine = [re.sub('@', '', i) for i in amine]

    for perm in itertools.product(cooh, amine):
        
        cooh_i = perm[0]
        amine_i = perm[1]
        
        smarts_products = react(rxn_smarts, perm)

        for p_1 in smarts_products:
            for p_2 in products:
                p_2 = re.sub('@', '', p_2)
                patt = Chem.MolFromSmiles(p_2)  
                if Chem.MolToInchiKey(p_1[0]) == Chem.MolToInchiKey(patt):  
                    return cooh_i, amine_i, p_2
    return None


def get_largest_mol(smiles, to_smiles=False):
    mol = Chem.MolFromSmiles(smiles)
    if mol is None:
        return
    mol_frags = rdmolops.GetMolFrags(mol, asMols=True)
    largest_mol = max(mol_frags, default=mol, key=lambda m: m.GetNumAtoms())
    if to_smiles:
        return mv.standardize_smiles(Chem.MolToSmiles(largest_mol))
    return largest_mol


def standardize_tautomer(mol, max_tautomers=1000):
    params = rdMolStandardize.CleanupParameters()
    params.maxTautomers = max_tautomers
    enumerator = rdMolStandardize.TautomerEnumerator(params)
    cm = enumerator.Canonicalize(mol)
    return cm


def reorder_tautomers(m):
    enumerator = rdMolStandardize.TautomerEnumerator()
    canon = enumerator.Canonicalize(m)
    csmi = Chem.MolToSmiles(canon)
    res = [canon]
    tauts = enumerator.Enumerate(m)
    smis = [Chem.MolToSmiles(x) for x in tauts]
    stpl = sorted(
        (x, y) for x, y in zip(smis, tauts) if x!=csmi
    )
    res += [y for _, y in stpl]
    return res


def rm_index(df):
    return df.loc[:, ~df.columns.str.match('Unnamed')]


def rm_col(df, col_name):
    return df.loc[:, ~df.columns.str.match(col_name)]


def shuffle_df(df):
    return df.sample(frac=1).reset_index(drop=True)



import re

import cirpy
import pubchempy as pcp
import molvs as mv
from psycopg import sql

from jupyfuncs.sql import connect_by_infofile


def query_from_cir(query_name: str):
    smiles = None
    # cas_list = []
    # name_list = [host=172.20.0.5 dbname=pistachio port=5432 user=postgres password=woshipostgres]

    cas_list = cirpy.resolve(query_name, 'cas')
    if cas_list is None or not cas_list:
        cas_list = []
    if isinstance(cas_list, str):
        cas_list = [cas_list]

    name_list = cirpy.resolve(query_name, 'names')
    if name_list is None or not name_list:
        name_list = []
    if isinstance(name_list, str):
        name_list = [name_list]

    smiles = cirpy.resolve(query_name, 'smiles')
    try:
        smiles = mv.standardize_smiles(smiles)
    except Exception as e:
        print(e)

    return smiles, cas_list, name_list


def query_from_pubchem(query_name: str):
    results = pcp.get_compounds(query_name, 'name')
    smiles = None
    name_list = set()
    cas_list = set()

    if any(results):
        try:
            smiles = mv.standardize_smiles(results[0].canonical_smiles)
        except Exception as e:
            smiles = results[0].canonical_smiles
            print(smiles)
            print(e)
        for compound in results:
            name_list.update(set(compound.synonyms))
            for syn in compound.synonyms:
                match = re.match('(\d{2,7}-\d\d-\d)', syn)
                if match:
                    cas_list.add(match.group(1))
        
        cas_list = list(cas_list)
        name_list = list(name_list)

    return smiles, cas_list, name_list


def query_a_compound(
    query_name: str,
    connect_info: str,
    by: str = 'name',
    log_file: str = './err.log'
):
    fei = None
    found = False  

    if by != 'smiles':
        query_name = query_name.lower()
    
    by = 'name'
    table = by + '_maps'
    # query_name = 'adipic acid'
    query = sql.SQL(
        "select fei from {table} where {by} = %s"
    ).format(
        table=sql.Identifier(table),
        by=sql.Identifier(by)
    )
    conn = connect_by_infofile(connect_info)
    
    cur = conn.execute(query, [query_name]).fetchone()

    if cur is not None:
        fei = cur[0]
        found = True
        return fei
 
    if not found:
        try:
            smiles, cas_list, name_list = query_from_pubchem(query_name) 
        except Exception as e:
            print(e)
            smiles, cas_list, name_list = None, [], []
        if smiles is not None:
            found = True
        else:
            try:
                smiles, cas_list, name_list = query_from_cir(query_name)
            except Exception as e:
                print(e)
                smiles, cas_list, name_list = None, [], []
            if smiles is not None:
                found = True
    
    if not found:
        with open(log_file, 'a') as f:
            f.write(query_name)
            f.write('\n')
        return
        # raise ValueError('给的啥破玩意儿查都查不着！')
    else:
        query_compound = sql.SQL(
            "select fei from compounds where smiles = %s"
        )
        cur = conn.execute(query_compound, [smiles]).fetchone()
        if cur is not None:
            fei = cur[0]
        elif any(cas_list):
            fei = cas_list[0]
            insert_compounds_sql = sql.SQL(
                "INSERT INTO compounds (fei, smiles) VALUES (%s, %s) ON CONFLICT (fei) DO NOTHING"
            )
            conn.execute(insert_compounds_sql, [fei, smiles])
        for cas in cas_list:
            insert_cas_map_sql = sql.SQL(
                "INSERT INTO cas_maps (fei, cas) VALUES (%s, %s) ON CONFLICT (cas) DO NOTHING"
            )
            try:
                conn.execute(insert_cas_map_sql, [fei, cas])
            except Exception as e:
                print(e)
        for name in name_list:
            insert_name_map_sql = sql.SQL(
                "INSERT INTO name_maps (fei, name) VALUES (%s, %s) ON CONFLICT (fei, name) DO NOTHING"
                # "INSERT INTO name_maps (fei, name) VALUES (%s, %s)"
            ) 
            try:
                conn.execute(insert_name_map_sql, [fei, name.lower()])
            except Exception as e:
                print(e)

    conn.commit()
    conn.close()

    return fei



from rdkit.Chem.Scaffolds import rdScaffoldNetwork
from rdkit import RDLogger
RDLogger.DisableLog('rdApp.info')


def create_sn(
    mols,
    includeGenericScaffolds=False,
    includeGenericBondScaffolds=False,
    includeScaffoldsWithAttachments=True,
    includeScaffoldsWithoutAttachments=False,
    pruneBeforeFragmenting=True,
    keepOnlyFirstFragment=True
):
    RDLogger.DisableLog('rdApp.info')
    scaffParams = rdScaffoldNetwork.ScaffoldNetworkParams()
    scaffParams.collectMolCounts = True
    scaffParams.includeGenericScaffolds = includeGenericScaffolds
    scaffParams.includeScaffoldsWithoutAttachments = includeScaffoldsWithoutAttachments
    scaffParams.keepOnlyFirstFragment = keepOnlyFirstFragment
    scaffParams.includeGenericBondScaffolds = includeGenericBondScaffolds
    scaffParams.includeScaffoldsWithAttachments = includeScaffoldsWithAttachments
    scaffParams.pruneBeforeFragmenting = pruneBeforeFragmenting
    net = rdScaffoldNetwork.CreateScaffoldNetwork(mols, scaffParams)
    return net


from rdkit import Chem
import numpy as np
from rdkit.Chem import AllChem
from rdkit.Chem import MACCSkeys


__all__ = [
    'get_fp',
]


def get_rdnorm_fp(smiles):
    from descriptastorus.descriptors import rdNormalizedDescriptors
    generator = rdNormalizedDescriptors.RDKit2DNormalized()
    features = generator.process(smiles)[1:]
    arr = np.array(features)
    return arr


def get_maccs_fp(smiles):
    arr = np.zeros(167)
    try:
        mol = Chem.MolFromSmiles(smiles)
        vec = MACCSkeys.GenMACCSKeys(mol)
        bv = list(vec.GetOnBits())
        arr[bv] = 1
    except Exception as e:
        print(e)
    return arr


def get_morgan_fp(smiles):
    mol = Chem.MolFromSmiles(smiles)
    vec = AllChem.GetMorganFingerprintAsBitVect(mol, 2, nBits=1024)
    bv = list(vec.GetOnBits())
    arr = np.zeros(1024)
    arr[bv] = 1
    return arr


fp_dict = {
    'rdnorm': get_rdnorm_fp,
    'maccs': get_maccs_fp,
    'morgan': get_morgan_fp
}


def get_fp(smiles, fp='maccs'):
    return fp_dict[fp](smiles)


def list_diff(listA, listB, mode="intersection"):
    if mode == "intersection":
        ret = list(set(listA).intersection(set(listB)))
    elif mode == "union":
        ret = list(set(listA).union(set(listB)))
    elif mode == "diff":
        ret = list(set(listB).difference(set(listA)))
    
    return ret


from setuptools import setup, find_packages
import setuptools_scm

def read_requirements():
    """读取 requirements.txt 文件并返回依赖列表"""
    with open('requirements.txt', 'r', encoding='utf-8') as file:
        return [
            line.strip()
            for line in file
            if line.strip() and not line.startswith('#')
        ]


def custom_version_scheme(version):
    """自定义版本号方案，确保没有 .dev 后缀"""
    if version.exact:
        return version.format_with("{tag}")
    elif version.distance:
        return f"{version.format_next_version()}.post{version.distance}"
    else:
        return version.format_with("0.0.0")

def custom_local_scheme(version):
    """自定义本地版本方案，确保没有本地版本后缀"""
    return ""

setup(
    name="jupyfuncs",
    use_scm_version={
        "version_scheme": custom_version_scheme,
        "local_scheme": custom_local_scheme,
        "write_to": "jupyfuncs/_version.py"
    },
    author="Jianxing Hu",
    author_email="j.hu@pku.edu.cn",
    description="A collection of functions for Jupyter notebooks",
    long_description=open('README.md', 'r', encoding='utf-8').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/huluxiaohuowa/jupyfuncs",
    packages=find_packages(),
    package_data={
        'jupyfuncs': [
            'datasets/*.*',
            'datasets/*/*.*',
            'defined_BaseFeatures.fdef'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    setup_requires=['setuptools_scm'],
    install_requires=read_requirements()
)



from openai import OpenAI
# import traceback


def chat_oai_stream(
    base_url="http://127.0.0.1:8000/v1",
    api_key="dummy_key",
    model="/data/models/Qwen-7B-Chat-Int4",
    prompt="Who are you?",
):
    """Chat with OpenAI's GPT-3 model using the specified parameters.
    
    Args:
        base_url (str): The base URL for the OpenAI API. Default is "http://127.0.0.1:8000/v1".
        api_key (str): The API key for accessing the OpenAI API. Default is "dummy_key".
        model (str): The model ID to use for the chat. Default is "/data/models/Qwen-7B-Chat-Int4".
        prompt (str): The initial prompt for the chat conversation.
    
    Yields:
        str: The generated content from the chat conversation.
    
    """
    client = OpenAI(
        base_url=base_url,
        api_key=api_key,
    )
    response = client.chat.completions.create(
        model=model,
        messages=[{
            "role": "user",
            "content": prompt
        }],
        stream=True
    )
    
    for chunk in response:
        content = chunk.choices[0].delta.content
        yield content


def chat_oai_invoke(
    base_url="http://127.0.0.1:8000/v1",
    api_key="dummy_key",
    model="/data/models/Qwen-7B-Chat-Int4",
    prompt="Who are you?",
):
    """Invoke OpenAI chat API to generate a response based on the given prompt.
    
    Args:
        base_url (str): The base URL of the OpenAI API. Default is "http://127.0.0.1:8000/v1".
        api_key (str): The API key for accessing the OpenAI API. Default is "dummy_key".
        model (str): The model to use for generating the response. Default is "/data/models/Qwen-7B-Chat-Int4".
        prompt (str): The prompt message to start the conversation. Default is "Who are you?".
    
    Returns:
        str: The response generated by the OpenAI chat API based on the prompt.
    """
    client = OpenAI(
        base_url=base_url,
        api_key=api_key,
    )
    response = client.chat.completions.create(
        model=model,
        messages=[{
            "role": "user",
            "content": prompt
        }],
        stream=False
    )
    
    return response.choices[0].message.content

# client = OpenAI(
#     base_url="http://127.0.0.1:8000/v1",
#     api_key="dummy_key"  # 使用虚拟的API Key
# )


# response = client.chat.completions.create(
#     model="/data/models/Qwen-7B-Chat-Int4",
#     messages=[{"role": "user", "content": "给我讲个鬼故事"}],
#     stream=True
# )
# for chunk in response:
#     content = chunk.choices[0].delta.content
#     if content:
#         print(content, end='', flush=True)
# print()


import functools
import logging

from rdkit import Chem
from rdkit.Chem import AllChem 


log = logging.getLogger(__name__)


__all__ = [
    "Normalizer",
    "Normalization",
    "NORMALIZATIONS",
]


def memoized_property(fget):
    """Decorator to create memoized properties."""
    attr_name = "_{}".format(fget.__name__)

    @functools.wraps(fget)
    def fget_memoized(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fget(self))
        return getattr(self, attr_name)

    return property(fget_memoized)


class Normalization(object):
    """A normalization transform defined by reaction SMARTS."""

    def __init__(self, name, transform):
        """
        :param string name: A name for this Normalization
        :param string transform: Reaction SMARTS to define the transformation.
        """
        log.debug("Initializing Normalization: %s", name)
        self.name = name
        self.transform_str = transform

    @memoized_property
    def transform(self):
        log.debug("Loading Normalization transform: %s", self.name)
        return AllChem.ReactionFromSmarts(self.transform_str)

    def __repr__(self):
        return "Normalization({!r}, {!r})".format(self.name, self.transform_str)

    def __str__(self):
        return self.name


NORMALIZATIONS = [
    Normalization(
        "Nitro to N+(O-)=O",
        "[*:1][N,P,As,Sb:2](=[O,S,Se,Te:3])=[O,S,Se,Te:4]>>[*:1][*+1:2]([*-1:3])=[*:4]",
    ),
    Normalization(
        "Sulfone to S(=O)(=O)", "[S+2:1]([O-:2])([O-:3])>>[S+0:1](=[O-0:2])(=[O-0:3])"
    ),
    Normalization("Pyridine oxide to n+O-", "[n:1]=[O:2]>>[n+:1][O-:2]"),
    Normalization(
        "Azide to N=N+=N-", "[*,H:1][N:2]=[N:3]#[N:4]>>[*,H:1][N:2]=[N+:3]=[N-:4]"
    ),
    Normalization("Diazo/azo to =N+=N-", "[*:1]=[N:2]#[N:3]>>[*:1]=[N+:2]=[N-:3]"),
    Normalization(
        "Sulfoxide to -S+(O-)-",
        "[!O:1][S+0;X3:2](=[O:3])[!O:4]>>[*:1][S+1:2]([O-:3])[*:4]",
    ),
    Normalization(
        "Phosphate to P(O-)=O",
        "[O,S,Se,Te;-1:1][P+;D4:2][O,S,Se,Te;-1:3]>>[*+0:1]=[P+0;D5:2][*-1:3]",
    ),
    Normalization(
        "Amidinium to C(=NH2+)NH2",
        "[C,S;X3+1:1]([NX3:2])[NX3!H0:3]>>[*+0:1]([N:2])=[N+:3]",
    ),
    Normalization(
        "Normalize hydrazine-diazonium",
        "[CX4:1][NX3H:2]-[NX3H:3][CX4:4][NX2+:5]#[NX1:6]>>[CX4:1][NH0:2]=[NH+:3][C:4][N+0:5]=[NH:6]",
    ),
    Normalization(
        "Recombine 1,3-separated charges",
        "[N,P,As,Sb,O,S,Se,Te;-1:1]-[A:2]=[N,P,As,Sb,O,S,Se,Te;+1:3]>>[*-0:1]=[*:2]-[*+0:3]",
    ),
    Normalization(
        "Recombine 1,3-separated charges",
        "[n,o,p,s;-1:1]:[a:2]=[N,O,P,S;+1:3]>>[*-0:1]:[*:2]-[*+0:3]",
    ),
    Normalization(
        "Recombine 1,3-separated charges",
        "[N,O,P,S;-1:1]-[a:2]:[n,o,p,s;+1:3]>>[*-0:1]=[*:2]:[*+0:3]",
    ),
    Normalization(
        "Recombine 1,5-separated charges",
        "[N,P,As,Sb,O,S,Se,Te;-1:1]-[A+0:2]=[A:3]-[A:4]=[N,P,As,Sb,O,S,Se,Te;+1:5]>>[*-0:1]=[*:2]-[*:3]=[*:4]-[*+0:5]",
    ),
    Normalization(
        "Recombine 1,5-separated charges",
        "[n,o,p,s;-1:1]:[a:2]:[a:3]:[c:4]=[N,O,P,S;+1:5]>>[*-0:1]:[*:2]:[*:3]:[c:4]-[*+0:5]",
    ),
    Normalization(
        "Recombine 1,5-separated charges",
        "[N,O,P,S;-1:1]-[c:2]:[a:3]:[a:4]:[n,o,p,s;+1:5]>>[*-0:1]=[c:2]:[*:3]:[*:4]:[*+0:5]",
    ),
    Normalization(
        "Normalize 1,3 conjugated cation",
        "[N,O;+0!H0:1]-[A:2]=[N!$(*[O-]),O;+1H0:3]>>[*+1:1]=[*:2]-[*+0:3]",
    ),
    Normalization(
        "Normalize 1,3 conjugated cation",
        "[n;+0!H0:1]:[c:2]=[N!$(*[O-]),O;+1H0:3]>>[*+1:1]:[*:2]-[*+0:3]",
    ),
    Normalization(
        "Normalize 1,3 conjugated cation",
        "[N,O;+0!H0:1]-[c:2]:[n!$(*[O-]),o;+1H0:3]>>[*+1:1]=[*:2]:[*+0:3]",
    ),
    Normalization(
        "Normalize 1,5 conjugated cation",
        "[N,O;+0!H0:1]-[A:2]=[A:3]-[A:4]=[N!$(*[O-]),O;+1H0:5]>>[*+1:1]=[*:2]-[*:3]=[*:4]-[*+0:5]",
    ),
    Normalization(
        "Normalize 1,5 conjugated cation",
        "[n;+0!H0:1]:[a:2]:[a:3]:[c:4]=[N!$(*[O-]),O;+1H0:5]>>[n+1:1]:[*:2]:[*:3]:[*:4]-[*+0:5]",
    ),
    Normalization(
        "Normalize 1,5 conjugated cation",
        "[N,O;+0!H0:1]-[c:2]:[a:3]:[a:4]:[n!$(*[O-]),o;+1H0:5]>>[*+1:1]=[c:2]:[*:3]:[*:4]:[*+0:5]",
    ),
    Normalization(
        "Normalize 1,5 conjugated cation",
        "[n;+0!H0:1]1:[a:2]:[a:3]:[a:4]:[n!$(*[O-]);+1H0:5]1>>[n+1:1]1:[*:2]:[*:3]:[*:4]:[n+0:5]1",
    ),
    Normalization(
        "Normalize 1,5 conjugated cation",
        "[n;+0!H0:1]:[a:2]:[a:3]:[a:4]:[n!$(*[O-]);+1H0:5]>>[n+1:1]:[*:2]:[*:3]:[*:4]:[n+0:5]",
    ),
    Normalization(
        "Charge normalization",
        "[F,Cl,Br,I,At;-1:1]=[O:2]>>[*-0:1][O-:2]"),
    Normalization(
        "Charge recombination", "[N,P,As,Sb;-1:1]=[C+;v3:2]>>[*+0:1]#[C+0:2]"
    ),
    Normalization(
        "Nitro to N+(O-)=O",
        "[N;X3:1](=[O:2])=[O:3]>>[*+1:1]([*-1:2])=[*:3]"),
    Normalization(
        "Diazonium N",
        "[*:1]-[N;X2:2]#[N;X1:3]>>[*:1]-[*+1:2]#[*:3]",
    ),
    Normalization(
        "Quaternary N",
        "[N;X4;v4;+0:1]>>[*+1:1]",
    ),
    Normalization(
        "Trivalent O",
        "[*:1]=[O;X2;v3;+0:2]-[#6:3]>>[*:1]=[*+1:2]-[*:3]",
    ),
    Normalization(
        "Sulfoxide to -S+(O-)",
        "[!O:1][S+0;D3:2](=[O:3])[!O:4]>>[*:1][S+1:2]([O-:3])[*:4]",
    ),
    Normalization(
        "Sulfoxide to -S+(O-) 2",
        "[!O:1][SH1+1;D3:2](=[O:3])[!O:4]>>[*:1][S+1:2]([O-:3])[*:4]",
    ),
    Normalization(
        "Trivalent S",
        "[O:1]=[S;D2;+0:2]-[#6:3]>>[*:1]=[*+1:2]-[*:3]",
    ),

    Normalization(
        "Bad amide tautomer1",
        "[C:1]([OH1;D1:2])=;!@[NH1:3]>>[C:1](=[OH0:2])-[NH2:3]",
    ),
    Normalization(
        "Bad amide tautomer2",
        "[C:1]([OH1;D1:2])=;!@[NH0:3]>>[C:1](=[OH0:2])-[NH1:3]",
    ),
    Normalization(
        "Halogen with no neighbors", "[F,Cl,Br,I;X0;+0:1]>>[*-1:1]",
    ),
    Normalization(
        "Odd pyridine/pyridazine oxide structure",
        "[C,N;-;D2,D3:1]-[N+2;D3:2]-[O-;D1:3]>>[*-0:1]=[*+1:2]-[*-:3]",
    ),
    Normalization(
        "qunimade2",
        "[n&H0:1][n&H1:2][n&H1,c;R2:3][c&H1,n&H1:4][c,n&H1:5](=[S,N,O:7])[n&H1:6]>>[n&H0:1][n&H1:2][n&H0,c;R2:3][c&H1,n&H0:4][c,n&H0:5]([S,N,O:7])[n&H0:6]"
    ),
    Normalization(
        "qunimade",
        "[c,n&H0,n&H1:2][n&H0,n&H1,c:3][c,n&H0,n&H1:4][c,n&H0,n&H1:5](=[S,N,O:1])>>[c,n&H0:2][n&H0,c:3][c,n&H0:4][c,n&H0:5]([S,N,O:1])"
    ),
]


class Normalizer(object):
    """A class for applying Normalization transforms.
    This class is typically used to apply a series of Normalization transforms to correct functional groups and
    recombine charges. Each transform is repeatedly applied until no further changes occur.
    """

    def __init__(self, normalizations=NORMALIZATIONS):
        """Initialize a Normalizer with an optional custom list of :class:`~molvs.normalize.Normalization` transforms.
        :param normalizations: A list of  :class:`~molvs.normalize.Normalization` transforms to apply.
        :param int max_restarts: The maximum number of times to attempt to apply the series of normalizations (default
                                 200).
        """
        log.debug("Initializing Normalizer")
        self.normalizations = normalizations

    def __call__(self, mol):
        """Calling a Normalizer instance like a function is the same as calling its normalize(mol) method."""
        return self.normalize(mol)

    def normalize(self, mol):
        """Apply a series of Normalization transforms to correct functional groups and recombine charges.
        A series of transforms are applied to the molecule. For each Normalization, the transform is applied repeatedly
        until no further changes occur. If any changes occurred, we go back and start from the first Normalization
        again, in case the changes mean an earlier transform is now applicable. The molecule is returned once the entire
        series of Normalizations cause no further changes or if max_restarts (default 200) is reached.
        :param mol: The molecule to normalize.
        :type mol: :rdkit:`Mol <Chem.rdchem.Mol-class.html>`
        :return: The normalized fragment.
        :rtype: :rdkit:`Mol <Chem.rdchem.Mol-class.html>`
        """
        log.debug("Running Normalizer")
        # Normalize each fragment separately to get around quirky RunReactants behaviour
        fragments = []
        for fragment in Chem.GetMolFrags(mol, asMols=True):
            fragments.append(self._normalize_fragment(fragment))
        # Join normalized fragments into a single molecule again
        outmol = fragments.pop()
        for fragment in fragments:
            outmol = Chem.CombineMols(outmol, fragment)
        Chem.SanitizeMol(outmol)
        return outmol

    def _normalize_fragment(self, mol):
        for normalization in self.normalizations:
            product = self._apply_transform(mol, normalization.transform)
            if product:
                mol = product
        return mol

    def _apply_transform(self, mol, rule):
        """Repeatedly apply normalization transform to molecule until no changes occur.
        It is possible for multiple products to be produced when a rule is applied. The rule is applied repeatedly to
        each of the products, until no further changes occur or after 20 attempts. If there are multiple unique products
        after the final application, the first product (sorted alphabetically by SMILES) is chosen.
        """
        mols = [mol]
        for n in range(20):
            products = {}
            for mol in mols:
                for product in [x[0] for x in rule.RunReactants((mol,))]:
                    if Chem.SanitizeMol(product, catchErrors=True) == 0:
                        products[
                            Chem.MolToSmiles(product, isomericSmiles=True)
                        ] = product
            if products:
                mols = [products[s] for s in sorted(products)]
            else:
                # If n == 0, the rule was not applicable and we return None
                return mols[0] if n > 0 else None


import psycopg


def connect_by_infofile(info_file: str) -> psycopg.Connection:
    """Create a postgres connection

    Args:
        info_file (str): 
            the path of the connection info like
            host=127.0.0.1 dbname=dbname port=5432 user=postgres password=lala

    Returns:
        psycopg.Connection: 
            the connection instance should be closed after committing.
    """
    conn = psycopg.connect(
        open(info_file).readline()
    )
    return conn
# with psycopg.connect(
#     open('./conn.info').readline()
# ) as conn:
#     cur = conn.execute('select * from name_maps;')
#     cur.fetchone()
#     for record in cur:
#         print(record)
#     conn.commit()
#     conn.close()


import requests


def get_proxies(
    pool_server='http://172.20.0.9:5010',
    https=False
):
    http_proxy = requests.get(f"{pool_server}/get/").json().get("proxy")
    https_proxy = requests.get(f"{pool_server}/get/?type=https").json().get("proxy")
    if https:
        proxy_dict = {
            'http': f'http://{https_proxy}',
            'https': f'https://{https_proxy}'
        }
    else:
        proxy_dict = {
            'http': f'http://{http_proxy}',
            'https': f'https://{http_proxy}'
        }
    return proxy_dict